import paddle
import paddle.nn as nn


class WeightQuantizationLinear(nn.Layer):
    """
    Args:
        linear (nn.Linear): 线性层.
    Returns:
        None.
    """
    def __init__(
            self, linear):
        """
        从父类中继承并重写__init__()方法。
        
        Args:
            linear (torch.nn.Linear): 线性层的实例。
        
        """
        super().__init__()
        self.linear = linear
        self.in_features = linear.weight.shape[0]
        self.out_features = linear.weight.shape[1]
        self.weight_name = linear.weight.name
        self.quant_weight_name = ".".join([self.weight_name, "quant_weight"])

    def forward(self, x):
        """
        Args:
            x (Tensor): input tensor with shape of (N, C, H, W).
        
        Returns:
            Tensor: output tensor with the same shape as `x`.
        
        Raises:
            NotImplementedError: if this method is not implemented in the subclass.
        """
        raise NotImplementedError()

    def quantize(self, weight):
        """
        Args:
            weight (torch.Tensor): Input tensor with shape :math:`[C_{in}, C_{out}]`.
        
        Returns:
            torch.Tensor: Quantized weights with the same type as ``weight``.
        
        Raises:
            NotImplementedError: When this method is called directly. Subclasses should implement it.
        """
        raise NotImplementedError()


class Linear4bit(WeightQuantizationLinear):
    """
    Args:
        linear (nn.Linear): 线性层.
        quant_type (str): 量化类型，目前支持"nf4"和"fp4"。

    Returns:
        None.
    """
    def __init__(
            self,
            linear, quant_type="nf4"):
        """
        将线性层封装为具有4位量化的线性层。
        
        Args:
            linear (paddle.nn.Linear): 需要被量化的线性层对象。
        
        Returns:
            Linear4bit: 返回具有4位量化的线性层对象。
        
        """
        super(Linear4bit, self).__init__(linear)
        self.quant_dtype = "int4"
        self.weight_dtype = "uint8"
        self.quant_scale_suffix = "_quant_scale"
        self.double_quant_scale_suffix = "_double_quant_scale"
    
        self.quant_weight = self.create_parameter(
            shape=[self.out_features // 2, self.in_features],
            attr=paddle.nn.initializer.Constant(value=0),
            dtype=self.weight_dtype,
            is_bias=False, )

    def forward(self, x):
        """
        Args:
            x (Tensor): input tensor with shape of (N, C, H, W).
        
        Returns:
            Tensor: output tensor with the same shape as `x`.
        
        Raises:
            NotImplementedError: if this method is not implemented in the subclass.
        """
        raise NotImplementedError()

    def quantize(self, weight):
        """
        Args:
            weight (torch.Tensor): Input tensor with shape :math:`[C_{in}, C_{out}]`.
        
        Returns:
            torch.Tensor: Quantized weights with the same type as ``weight``.
        
        Raises:
            NotImplementedError: When this method is called directly. Subclasses should implement it.
        """
        raise NotImplementedError()
