import paddle
import paddle.nn as nn
from paddleslim.lc.quantizers import NF4Quantizer
from .linear import Linear4bit


class NF4Linear(Linear4bit):
    """
    Args:
        linear (nn.Linear): 线性层.
        block_size (int, optional): 块大小. 默认值是64.
        use_double_quant (bool, optional): 是否使用双量化. 默认值是False.

    Returns:
        None.
    """
    def __init__(
            self,
            linear,
            block_size=64,
            use_double_quant=False, ):
        """初始化 NF4 线性层。
        
        Args:
            linear (torch.nn.Linear): PyTorch 的 Linear 层对象。
            block_size (int, optional): 分块大小，默认值为 64。
            use_double_quant (bool, optional): 是否使用双精度量化，默认值为 False。
        
        Returns:
            None
        """

        super(NF4Linear, self).__init__(linear, quant_type="nf4")
        self.block_size = block_size
        self.double_quant = use_double_quant
        self.quantizer = NF4Quantizer(block_size, self.double_quant)

    def quantize(self):
        """将模型中所有Linear层的权重量化为两半大小的量化矩阵。
        
        Args:
            无参数。
        
        Returns:
            无返回值，执行完此操作后，模型中的Linear层的权重已被转换为量化矩阵。
        """
        tmp_weight = paddle.to_tensor(self.linear.weight, place=paddle.device.get_device())
        quantized_weight = self.quantizer.quantize(tmp_weight).reshape([self.out_features // 2, self.in_features])
        paddle.assign(quantized_weight, self.quant_weight)
        del self.linear.weight, tmp_weight, quantized_weight

    def forward(self, x):
        """
        Args:
            x (Tensor): shape = [batch_size * seq_len, input_dim]
        
        Returns:
            Tensor: shape = [batch_size * seq_len, output_dim]
        
        """
        return self.quantizer.matmul(x, self.quant_weight.reshape([-1, 1]), self.linear.bias)


class NF4RowParallelLinear(NF4Linear):
    """"""
    def __init__(
            self,
            linear,
            block_size=64,
            use_double_quant=False, ):
        """
        构造函数，初始化NF4RowParallelLinear类
        
        Args:
            linear (torch.nn.Linear): 一个PyTorch线性层实例
            block_size (int): 表示分块的大小，默认值为64
            use_double_quant (bool): 是否使用双重量化，默认值为False
        
        Returns:
            NoneType: 返回None，此处不需要返回任何对象
        
        """
        super(NF4RowParallelLinear, self).__init__(linear, block_size, use_double_quant)
        self.block_size = block_size
        self.double_quant = use_double_quant
        self.quantizer = NF4Quantizer(block_size, self.double_quant)
        self.input_is_parallel = linear.input_is_parallel
        self.is_mp = linear.is_mp
        self.model_parallel_group = linear.model_parallel_group

    def forward(self, input):
        """
        将模型分散并行化。
        
        Args:
            input (Tensor): 需要计算的输入张量。
        
        Returns:
            Tensor: 计算完成后的输出张量。
        """
        if self.input_is_parallel or (not self.is_mp):
            input_parallel = input
        else:
            # split last dim
            input_parallel = paddle.distributed.collective._c_split(
                input, group=self.model_parallel_group)

        if self.is_mp:
            output_parallel = self.quantizer.matmul(input_parallel, self.quant_weight.reshape([-1, 1]), None)
            output_ = paddle.distributed.collective._mp_allreduce(
                output_parallel,
                group=self.model_parallel_group,
                use_calc_stream=True,
                use_model_parallel=True)
            output = output_ + self.linear.bias if self.linear.bias is not None else output_
        else:
            output = self.quantizer.matmul(input_parallel, self.quant_weight.reshape([-1, 1]), self.linear.bias)
        return output

class NF4ColumnParallelLinear(NF4Linear):
    """"""
    def __init__(
            self,
            linear,
            block_size=64,
            use_double_quant=False, ):
        """
        """
        super(NF4ColumnParallelLinear, self).__init__(linear, block_size, use_double_quant)
        self.block_size = block_size
        self.double_quant = use_double_quant
        self.quantizer = NF4Quantizer(block_size, self.double_quant)
        self.is_mp = linear.is_mp
        self.model_parallel_group = linear.model_parallel_group
        self.gather_output = linear.gather_output

    def forward(self, input):
        """
        Args:
            input (Tensor): Input Tensor to be quantized.
        
        Returns:
            Tensor: Output Tensor of quantization.
        """
        if self.is_mp:
            input_parallel = paddle.distributed.collective._c_identity(
                input, group=self.model_parallel_group)
        else:
            input_parallel = input

        output_parallel = self.quantizer.matmul(input_parallel, self.quant_weight.reshape([-1, 1]), self.linear.bias)

        if self.gather_output and self.is_mp:
            output = paddle.distributed.collective._c_concat(
                output_parallel, group=self.model_parallel_group)
        else:
            output = output_parallel
        return output