# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from .base_quantizer import BaseQuantizer
from .quant_func import quantize_fp4, dequantize_fp4, quantize_8bit, dequantize_8bit


class FP4Quantizer(BaseQuantizer):

    def __init__(self, block_size=64, double_quant=False):
        """
        初始化FP4量化器（FP4Quantizer）
        
        Args:
            block_size (int, optional): 块大小，默认为64。
            double_quant (bool, optional): 是否使用双量化，默认为False。
        
        Attributes:
            block_size (int): 块大小。
            double_quant (bool): 是否使用双量化。
        
        """
        super(FP4Quantizer, self).__init__()
        self.block_size = block_size
        self.double_quant = double_quant

    def quantize(self, x):
        """Quantize the input tensor using fp32 to int8 quantization
        
        Args:
            x (torch.Tensor): Input Tensor of size :math:`B \\times C_{in}
                \\times H \\times W` to be quantized.
        
        Returns:
            torch.Tensor: Quantized output tensor in shape :math:`B \\times
                C_{out} \\times H \\times W`, where :math:`C_{out}` is the number
                of channels after quantization.
        
            tuple: Tuple with two elements, quant scale and offset, respectively
                for the input tensor quantized into 8 bit integer.
        
        """
        out, quant_scale = quantize_fp4(x, self.block_size)
        if self.double_quant:
            offset = quant_scale.mean()
            quant_scale -= offset
            qquant_scale, double_quant_scale = quantize_8bit(quant_scale, None, self.block_size, quant_type="dynamic_fp8")
            self.state = [qquant_scale, double_quant_scale, offset]
        else:
            self.state = quant_scale
        return out

    def dequantize(self, x):
        """
        根据模型的状态，对输入张量进行反量化操作。
        
        Args:
            x (paddle.Tensor): 输入张量，shape为[batch_size, num_channels, height, width]。
        
        Returns:
            paddle.Tensor: 经过反量化处理后的张量，shape和x相同。
        
        """
        if self.double_quant:
            qquant_scale, double_quant_scale, offset = self.state
            quant_scale = dequantize_8bit(qquant_scale, None, double_quant_scale, self.block_size, quant_type="dynamic_fp8")
            quant_scale += offset
        else:
            quant_scale = self.state 

        out = dequantize_fp4(
            x, quant_scale, blocksize=self.block_size)
        return out

    def matmul(self, x, y, bias):
        """
        Args:
            x (paddle.Tensor): Input tensor of shape (batch_size, input_dim)
            y (paddle.Tensor): Weight tensor of shape (input_dim, output_dim)
            bias (paddle.Tensor): Bias tensor of shape (output_dim,)
        
        Returns:
            paddle.Tensor: Output tensor of shape (batch_size, output_dim)
        
        """
        return x @ self.dequantize(y).cast(x.dtype).reshape((x.shape[-1], -1)) + bias
