# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from .base_quantizer import BaseQuantizer
from .quant_func import quantize_nf4, dequantize_nf4, quantize_8bit, dequantize_8bit


class NF4Quantizer(BaseQuantizer):

    def __init__(self, block_size=64, double_quant=False):
        """
        初始化量化器类，继承自父类torch.nn.Module。
        
        Args:
            block_size (int, optional): 分块大小。默认值为64。
            double_quant (bool, optional): 是否使用双量化模式。默认值为False。
        
        Returns:
            无返回值。
        
        """
        super(NF4Quantizer, self).__init__()
        self.block_size = block_size
        self.double_quant = double_quant

    def quantize(self, x):
        """Quantizes the input Tensor into int8 using dynamic scale quantization.
        
        Args:
            x (paddle.Tensor): Input Tensor to be quantized.
        
        Returns:
            paddle.Tensor: Quantized Tensor with type as 'int8'.
        
        """
        out, quant_scale = quantize_nf4(x, self.block_size)
        if self.double_quant:
            offset = quant_scale.mean()
            quant_scale -= offset
            qquant_scale, double_quant_scale = quantize_8bit(quant_scale, None, self.block_size, quant_type="dynamic_fp8")
            self.state = [qquant_scale, double_quant_scale, offset]
        else:
            self.state = quant_scale
        return out

    def dequantize(self, x):
        """
        将量化后的张量恢复为正常的浮点数。
        
        Args:
            x (paddle.Tensor): 量化后的张量，shape为[batch_size, channel, height, width]。
        
        Returns:
            paddle.Tensor: 恢复后得到的原始浮点型张量。
        
        """
        if self.double_quant:
            qquant_scale, double_quant_scale, offset = self.state
            quant_scale = dequantize_8bit(qquant_scale, None, double_quant_scale, self.block_size, quant_type="dynamic_fp8")
            quant_scale += offset
        else:
            quant_scale = self.state 

        out = dequantize_nf4(
            x, quant_scale, blocksize=self.block_size)
        return out

    def matmul(self, x, y, bias=None):
        """
        计算输入矩阵之间的乘法操作，并进行量化和重塑，输出为二维的矩阵。
        
        Args:
            x (paddle.Tensor): 输入矩阵X，支持任意维度。
            y (paddle.Tensor): 输入矩阵Y，与X具有相同数量的列数。
            bias (paddle.Tensor|None): 可选的偏置项，默认值为None。
        
        Returns:
            paddle.Tensor: 返回经过量化和重塑后的结果矩阵。
        
        """
        if bias is not None:
            return x @ self.dequantize(y).cast(x.dtype).reshape((x.shape[-1], -1)) + bias
        else:
            return x @ self.dequantize(y).cast(x.dtype).reshape((x.shape[-1], -1))
