import paddle
import random
import numpy as np
from re import T
from operator import index


from paddleslim.quant.layers import ShiftSmoothCacheKVMatMul
from .smooth_scale_utils import SearchSmoothCacheKVOnePiece
from ...utils.log import logger

__all__ = ['CacheKVSmooth']

class CacheKVSmooth():
    def __init__(self,
                 model,
                 search_smooth_func=SearchSmoothCacheKVOnePiece(),
                 search_start_steps=1000,
                 use_ema=False,
                 dtype='float16'):

        '''
        1. find all (LN linear), and save their correspongding name in ln_linear_dict
            algo: forward once, record order, find each pair of (LN linear)
        2. add hook for all linears in ln_linear_dict
        3. in hook, sample abs max activation for each step and save it in scale_dict by linear name
        4. when sampling done, update ln weight and linear weight 
        '''
        self.model = model
        self.search_smooth_func = search_smooth_func
        self.step = 0
        self.search_start_steps = search_start_steps
        self.dtype = dtype
        self.x_act_input = {}
        self.y_scale_dict = {}
        self.y_act_input = {}
        self.smooth_scale_dict = {}
        self.act_input = {}

        self.model.eval()
        self.print_step = 0
        self._apply_hook()
        self.search = False

    def _apply_hook(self):
        self._forward_hook_list = []
        for _, sub_layer in self.model.named_sublayers():
            if type(sub_layer) == ShiftSmoothCacheKVMatMul:
                forward_pre_hook_handle = sub_layer.register_forward_pre_hook(self._forward_add_smooth_pre_hook)
                self._forward_hook_list.append(forward_pre_hook_handle)

    
    def _forward_add_smooth_pre_hook(self, layer, input):
        self._sample_cachekv_scale(layer, input)
        return input

    def _sample_cachekv_scale(self, layer, input):
        kv_name = layer.full_name()
        x = input[0] if type(input) == tuple else input
        y = input[1] if type(input) == tuple else input
        x.stop_gradient = True
        y.stop_gradient = True
        layer_num = int(kv_name.split('_')[-1])
        if layer_num % 2 == 0:
            smooth_qk, smooth_kv = True, False
        else:
            smooth_qk, smooth_kv = False, True

        # x = paddle.add(x, layer.shift_weight)

        if self.step >= self.search_start_steps and self.search:
            # collect samples for x and y
            s_y = y
            select_index = paddle.to_tensor(random.sample(range(s_y.shape[1]), min(10, s_y.shape[1])))
            s_y = paddle.index_select(s_y, index=select_index, axis=1)
            top_value, top_index = paddle.topk(y.cast("float32"), k=6, axis=1)
            s_y = paddle.concat([s_y, top_value.cast(self.dtype)], axis=1)

            if smooth_qk:
                s_x = x
                select_index = paddle.to_tensor(random.sample(range(s_x.shape[1]), min(10, s_x.shape[1])))
                s_x = paddle.index_select(s_x, index=select_index, axis=1)
                top_value, top_index = paddle.topk(x.cast("float32"), k=6, axis=1)
                s_x = paddle.concat([s_x, top_value.cast(self.dtype)], axis=1)
            elif smooth_kv:
                # sampling for axis 2
                s_x = x
                select_index = paddle.to_tensor(random.sample(range(s_x.shape[2]), min(10, s_x.shape[2])))
                s_x = paddle.index_select(s_x, index=select_index, axis=2)
                top_value, top_index = paddle.topk(x.cast("float32"), k=6, axis=2)
                s_x = paddle.concat([s_x, top_value.cast(self.dtype)], axis=2)

                # sampling for axis 3
                select_index = paddle.to_tensor(random.sample(range(s_x.shape[3]), min(10, s_x.shape[3])))
                s_x = paddle.index_select(s_x, index=select_index, axis=3)
                top_value, top_index = paddle.topk(s_x.cast("float32"), k=6, axis=3)
                s_x = paddle.concat([s_x, top_value.cast(self.dtype)], axis=3)
            
            if kv_name in self.y_act_input:
                self.y_act_input[kv_name] = paddle.concat([s_y, self.y_act_input[kv_name]], axis=1)
                if smooth_qk:
                    self.x_act_input[kv_name] = paddle.concat([s_x, self.x_act_input[kv_name]], axis=1)
                elif smooth_kv:
                    self.x_act_input[kv_name] = paddle.concat([s_x, self.x_act_input[kv_name]], axis=3)
            else:
                self.y_act_input[kv_name] = s_y
                self.x_act_input[kv_name] = s_x
        
        y_abs_max = y.abs().max(axis=1, keepdim=True)
        if kv_name not in self.y_scale_dict:
            self.y_scale_dict[kv_name] = y_abs_max
        else:
            tmp1 = paddle.concat([y_abs_max, self.y_scale_dict[kv_name]], axis=0)
            self.y_scale_dict[kv_name] = tmp1.max(axis=0, keepdim=True)

        # per step print once
        # if self.print_step == self.step:
        #     print('[Smooth] Step [{}]: {}. abs_min: {}, abs_max: {}'.format(self.step, kv_name, round(float(self.y_scale_dict[kv_name].cast("float32").min()), 5), float(self.y_scale_dict[kv_name].max().cast("float32"))))

    def update_weight(self):
        for _, sub_layer in self.model.named_sublayers():
            if type(sub_layer) == ShiftSmoothCacheKVMatMul:
                layer_name = sub_layer.full_name()
                layer_num = int(layer_name.split('_')[-1])
                y_abs_max = self.y_scale_dict[layer_name].cast("float32")
                y_act_input = self.y_act_input[layer_name].cast("float32")
                x_act_input = self.x_act_input[layer_name].cast("float32")
                perm = [1, 2, 0, 3] if sub_layer.sequence_parallel else [0, 2, 1, 3]
                if layer_num % 2 == 0:
                    s = self.search_smooth_func.search(layer_name, x_act_input, y_act_input, y_abs_max, perm_x=perm, perm_y=perm, transpose_y=True)
                else:
                    s = self.search_smooth_func.search(layer_name, x_act_input, y_act_input, y_abs_max, perm_y=perm, sequence_parallel=sub_layer.sequence_parallel)
                self.smooth_scale_dict[layer_name] = s
        
        # update ShiftSmoothCacheKVMatMul weight
        for _, sub_layer in self.model.named_sublayers():
            if type(sub_layer) == ShiftSmoothCacheKVMatMul:
                layer_name = sub_layer.full_name()
                smooth_scale = self.smooth_scale_dict[layer_name]
                logger.debug(f"[smooth CacheKV] param: {layer_name}, before weight, abs_max: {float(sub_layer.smooth_weight.abs().max().cast("float32"))}")
                sub_layer.convert_weight(smooth_weight=smooth_scale)
                logger.debug(f"[smooth CacheKV] param: {layer_name}, after weight, abs_max: {float(sub_layer.smooth_weight.abs().max().cast("float32"))}")

        self._remove_hook()
        paddle.device.cuda.empty_cache()
    
    def _remove_hook(self):
        for hook in self._forward_hook_list:
            hook.remove()
        self._forward_hook_list = []