# Copyright (c) 2023  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
GPTQ.
"""
import math
import time
import numpy as np
import paddle
import paddle.nn as nn
from paddle.distributed.fleet.meta_parallel import ColumnParallelLinear, RowParallelLinear

from .utils import compute_scales, fake_quant
from ...utils.log import logger

__all__ = ['GPTQ']


class GPTQ(nn.Layer):
    """
    GPTQ new version.
    """
    def __init__(self,
                 model,
                 quant_bits=4,
                 weight_quant_method='groupwise',
                 blocksize=128,
                 percdamp=.1,
                 groupsize=-1,
                 actorder=True):
        super(GPTQ, self).__init__()
        self.model = model
        quant_bits = (1 << (quant_bits - 1)) - 1
        self.quant_bits = paddle.to_tensor(quant_bits, dtype='float32')
        self.weight_quant_method = weight_quant_method
        self.blocksize = blocksize
        self.percdamp = percdamp
        self.groupsize = groupsize
        self.actorder = actorder
        self.sampled_hessian = {}
        self.sampled_num = {}
        self._apply_hook()

    def _apply_hook(self):
        self._forward_hook_list = []
        for _, sub_layer in self.model.named_sublayers():
            if type(sub_layer) in [
                    ColumnParallelLinear, RowParallelLinear, paddle.nn.Linear
            ]:
                forward_pre_hook_handle = sub_layer.register_forward_pre_hook(
                    self._forward_pre_hook)
                self._forward_hook_list.append(forward_pre_hook_handle)

    def _forward_pre_hook(self, layer, input):
        weight = layer.weight.t()
        columns = weight.shape[1]
        self._sample(input, layer.full_name(), columns)
        return input

    def _sample(self, input, layer_name, columns):
        inp = input[0] if type(input) == tuple else input
        inp = inp.cast('float32')
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)
        tmp = inp.shape[0]
        if len(inp.shape) == 3:
            inp = inp.reshape((-1, inp.shape[-1]))
        inp = inp.t()
        if layer_name not in self.sampled_hessian:
            self.sampled_hessian[layer_name] = paddle.zeros(
                    (columns, columns), dtype='float32')
            self.sampled_num[layer_name] = 0
        hessian = self.sampled_hessian[layer_name]
        hessian *= self.sampled_num[layer_name] / (self.sampled_num[layer_name] + tmp)
        self.sampled_num[layer_name] += tmp
        inp = math.sqrt(2 / self.sampled_num[layer_name]) * inp
        hessian += paddle.matmul(inp, inp.t())
        self.sampled_hessian[layer_name] = hessian
        del inp

    def fasterquant(self):
        """
        fasterquant
        """
        for _, sub_layer in self.model.named_sublayers():
            layer_name = sub_layer.full_name()
            if layer_name in self.sampled_hessian:
                logger.debug(f"fast quantize layer: {layer_name} {sub_layer.weight.shape}")
                H = self.sampled_hessian[layer_name]
                W = sub_layer.weight.t().cast('float32')
                columns = W.shape[1]
                weight_scale = compute_scales(W.t(), method=self.weight_quant_method, group_size=self.groupsize)

                weight_scale /= self.quant_bits
                tick = time.time()
                dead = paddle.where(paddle.diag(H) == 0)
                if not paddle.is_empty(dead):
                    H[dead, dead] = 1
                    W[:, dead] = 0
                else:
                    logger.debug(f"skip dead")
                del dead
                if self.actorder:
                    perm = paddle.argsort(paddle.diag(H), descending=True)
                    W = W.transpose((1, 0))
                    W = W[perm].transpose((1, 0))
                    H = H[perm].transpose((1, 0))
                    H = H[perm].transpose((1, 0))

                Losses = paddle.zeros_like(W)
                Q = paddle.zeros_like(W)

                damp = self.percdamp * paddle.mean(paddle.diag(H))
                diag = paddle.arange(columns)
                H[diag, diag] += damp
                # H = paddle.inverse(H)
                # H = paddle.linalg.cholesky(H, upper=True)
                try:
                    H = paddle.inverse(H)
                    H = paddle.linalg.cholesky(H, upper=True)
                except:
                    logger.debug('We skip GPTQ this layer now.')
                    logger.debug(
                        'If you want GPTQ this layer, please try setting damp_percent larger or increasing the number of samples.'
                    )
                    continue
                Hinv = H
                for i1 in range(0, columns, self.blocksize):
                    i2 = min(i1 + self.blocksize, columns)
                    count = i2 - i1
                    W1 = W[:, i1:i2]
                    Q1 = paddle.zeros_like(W1)
                    Err1 = paddle.zeros_like(W1)
                    Losses1 = paddle.zeros_like(W1)
                    Hinv1 = Hinv[i1:i2, i1:i2]

                    for i in range(count):
                        w = W1[:, i]
                        d = Hinv1[i, i]
                        if self.groupsize != -1:
                            if (i1 + i) % self.groupsize == 0:
                                weight_scale = compute_scales(
                                    W[:, (i1 + i):(i1 + i + self.groupsize)].t(),
                                    method=self.weight_quant_method, group_size=self.groupsize)
                                weight_scale = weight_scale[0]
                                weight_scale /= self.quant_bits
                        q = paddle.clip(
                            paddle.round(w / weight_scale), -self.quant_bits - 1,
                            self.quant_bits)
                        q = q * weight_scale
                        Q1[:, i] = q
                        Losses1[:, i] = (w - q) ** 2 / d ** 2

                        err1 = (w - q) / d
                        W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                        Err1[:, i] = err1
                        del w, d, q, err1
                        paddle.device.cuda.empty_cache()

                    Q[:, i1:i2] = Q1
                    Losses[:, i1:i2] = Losses1 / 2
                    del Q1, Losses1
                    if Hinv[i1:i2, i2:].shape[1] != 0:
                        W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])
                    del Err1, W1, Hinv1
                    paddle.device.cuda.empty_cache()

                logger.debug(f"time {time.time() - tick}")
                logger.debug(f"error {paddle.sum(Losses).item()}")

                if self.actorder:
                    invperm = paddle.argsort(perm)
                    Q = Q.transpose((1, 0))
                    Q = Q[invperm].transpose((1, 0))
                    del invperm, perm
                paddle.device.cuda.empty_cache()
                param = sub_layer.weight
                Q = Q.t().cast(sub_layer.weight.dtype)
                paddle.assign(Q, output=param)
                del H, Q, Hinv, W, Losses
                paddle.device.cuda.empty_cache()

        self._remove_hook()
        paddle.device.cuda.empty_cache()

    def _remove_hook(self):
        for hook in self._forward_hook_list:
            hook.remove()
        self._forward_hook_list = [] 

