# Copyright (c) 2023  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
GPTQ.
"""
import math
import time
import numpy as np
import paddle
import paddle.nn as nn
from paddle.distributed.fleet.meta_parallel import ColumnParallelLinear, RowParallelLinear

from .utils import compute_scales
from ...utils.log import logger
__all__ = ['GPTQ']


class GPTQ(nn.Layer):
    """
    GPTQ old version.
    """
    def __init__(self,
                 layer,
                 quant_bits=4,
                 weight_quant_method='abs_max_channel_wise'):
        '''
        The implementation of GPTQ(https://arxiv.org/abs/2210.17323). 
        The codes here are based on https://github.com/IST-DASLab/gptq.
        
        Args:
        layer (paddle.nn.Layer): Layer object.
        quant_bits (int, optional): Number of bits to quantize the weight. Default: 4.
        weight_quant_method (str, optional): Method of weight quantization. Choosen from abs_max, abs_max_channel_wise and avg. Default: abs_max_channel_wise.

        Examples:
        .. code-block:: python
        
        from paddleslim.quant.advanced import GPTQ
        for cur_name, cur_layer in model.named_sublayers():
            if type(cur_layer) == paddle.nn.Linear:
                gptq_layer = GPTQ(cur_layer)
                # sample data
                for data in dataloader():
                    model(data)
                # quant weight
                gptq_layer.fasterquant()  
        '''
        super(GPTQ, self).__init__()
        self.layer = layer
        assert hasattr(layer,
                       'weight'), "Layer {} has no attribute 'weight'".format(
                           layer.full_name())
        assert type(self.layer) in [
            nn.Linear, ColumnParallelLinear, RowParallelLinear
        ], "Currently, GPTQ only supports linear layer and ColumnParallelLinear/RowParallelLinear layer"

        weight = layer.weight.t()

        self.rows = weight.shape[0]
        self.columns = weight.shape[1]
        self.hessian = paddle.zeros(
            (self.columns, self.columns), dtype='float32')
        self.nsamples = 0
        self.quantized = False
        self.weight_quant_method = weight_quant_method
        self.quant_bits = (1 << (quant_bits - 1)) - 1

    def forward(self, input):
        """
        forward
        """
        if not self.quantized:
            inp = input[0] if type(input) == tuple else input
            inp = inp.cast('float32')
            if len(inp.shape) == 2:
                inp = inp.unsqueeze(0)
            tmp = inp.shape[0]
            if type(self.layer) in [
                    nn.Linear, ColumnParallelLinear, RowParallelLinear
            ]:
                if len(inp.shape) == 3:
                    inp = inp.reshape((-1, inp.shape[-1]))
                inp = inp.t()

            self.hessian *= self.nsamples / (self.nsamples + tmp)
            self.nsamples += tmp

            inp = math.sqrt(2 / self.nsamples) * inp
            self.hessian += paddle.matmul(inp, inp.t())
            del inp
        return self.layer(input)

    def fasterquant(self,
                    blocksize=128,
                    percdamp=.01,
                    groupsize=-1,
                    actorder=True):
        """
        fasterquant
        """
        logger.debug(f"quant {self.layer.full_name()}")
        W = self.layer.weight.t().cast('float32')
        weight_scale = compute_scales(W.t(), method=self.weight_quant_method)
        weight_scale /= self.quant_bits
        tick = time.time()

        H = self.hessian
        del self.hessian
        dead = paddle.where(paddle.diag(H) == 0)
        H[dead, dead] = 1
        W[:, dead] = 0
        del dead
        if actorder:
            perm = paddle.argsort(paddle.diag(H), descending=True)
            W = W.transpose((1, 0))
            W = W[perm].transpose((1, 0))
            H = H[perm].transpose((1, 0))
            H = H[perm].transpose((1, 0))

        Losses = paddle.zeros_like(W)
        Q = paddle.zeros_like(W)

        damp = percdamp * paddle.mean(paddle.diag(H))
        diag = paddle.arange(self.columns)
        H[diag, diag] += damp

        H = paddle.inverse(H)
        H = paddle.linalg.cholesky(H, upper=True)
        Hinv = H

        for i1 in range(0, self.columns, blocksize):
            i2 = min(i1 + blocksize, self.columns)
            count = i2 - i1
            W1 = W[:, i1:i2]
            Q1 = paddle.zeros_like(W1)
            Err1 = paddle.zeros_like(W1)
            Losses1 = paddle.zeros_like(W1)
            Hinv1 = Hinv[i1:i2, i1:i2]

            for i in range(count):
                w = W1[:, i]
                d = Hinv1[i, i]
                if groupsize != -1:
                    if (i1 + i) % groupsize == 0:
                        weight_scale = compute_scales(
                            W[:, (i1 + i):(i1 + i + groupsize)].t(),
                            method=self.weight_quant_method)
                        weight_scale /= self.quant_bits

                q = paddle.clip(
                    paddle.round(w / weight_scale), -self.quant_bits - 1,
                    self.quant_bits)
                q = q * weight_scale
                Q1[:, i] = q
                Losses1[:, i] = (w - q) ** 2 / d ** 2

                err1 = (w - q) / d
                W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
                Err1[:, i] = err1
                del w, d, q, err1
                paddle.device.cuda.empty_cache()

            Q[:, i1:i2] = Q1
            Losses[:, i1:i2] = Losses1 / 2
            del Q1, Losses1
            if Hinv[i1:i2, i2:].shape[1] != 0:
                W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])
            del Err1, W1, Hinv1
            paddle.device.cuda.empty_cache()

        logger.debug(f"time {time.time() - tick}")
        logger.debug(f"error {paddle.sum(Losses).item()}")

        if actorder:
            invperm = paddle.argsort(perm)
            Q = Q.transpose((1, 0))
            Q = Q[invperm].transpose((1, 0))
            del invperm, perm

        param = self.layer.weight
        Q = Q.t().cast(self.layer.weight.dtype)
        paddle.assign(Q, output=param)

        self.quantized = True
        del H, Q, Hinv, W, Losses
        paddle.device.cuda.empty_cache()


# import math
# import time
# import numpy as np
# import torch
# import paddle
# import paddle.nn as nn
# from paddle.distributed.fleet.meta_parallel import ColumnParallelLinear, RowParallelLinear
# __all__ = ['GPTQ']

# class GPTQ(paddle.nn.Layer):
#     def __init__(self, layer):
#         super(GPTQ, self).__init__()
#         self.layer = layer
#         W = layer.weight.t()
#         if isinstance(self.layer, nn.Conv2D):
#             W = W.flatten(1)
        
#         self.rows = W.shape[0]
#         self.columns = W.shape[1]
#         self.H = paddle.zeros((self.columns, self.columns), dtype='float32')
#         self.nsamples = 0
#         self.quantized = False
#         quant_bits = 4
#         self.quant_bits = (1 << (quant_bits - 1)) - 1
#         self.weight_quant_type = 'abs_max_channel_wise'
        
#     def forward(self, input):
#         if not self.quantized:
#             inp = input[0] if type(input) == tuple else input
#             inp = inp.cast('float32')
#             # print(inp.shape)
#             if len(inp.shape) == 2:
#                 inp = inp.unsqueeze(0)
#             tmp = 1
#             if isinstance(self.layer, nn.Linear) or isinstance(self.layer, ColumnParallelLinear) or isinstance(self.layer, RowParallelLinear):
#                 if len(inp.shape) == 3:
#                     inp = inp.reshape((-1, inp.shape[-1]))
#                 inp = inp.t()
#             if isinstance(self.layer, nn.Conv2D):
#                 unfold = nn.Unfold(
#                     self.layer.kernel_size,
#                     dilation=self.layer.dilation,
#                     padding=self.layer.padding,
#                     stride=self.layer.stride
#                 )
#                 inp = unfold(inp)
#                 inp = inp.permute([1, 0, 2])
#                 inp = inp.flatten(1)

#             self.H *= self.nsamples / (self.nsamples + tmp)
#             self.nsamples += tmp

#             inp = math.sqrt(2 / self.nsamples) * inp
#             self.H += paddle.matmul(inp, inp.t())
#             del inp
#         return self.layer(input)

#     def fasterquant(
#         self, blocksize=128, percdamp=.01, groupsize=-1, actorder=False
#     ):
#         W = self.layer.weight.t().cast('float32')
#         W = np.array(W, dtype=np.float32)
#         W = torch.tensor(W, dtype=torch.float32)

#         if groupsize == -1:
#             weight_scale = compute_scales(W.t(), method=self.weight_quant_type)
#             weight_scale /= self.quant_bits
        
#         tick = time.time()

#         H = self.H
#         del self.H
#         H = np.array(H, dtype=np.float32)
#         H = torch.tensor(H, dtype=torch.float32)

#         dead = torch.diag(H) == 0
#         H[dead, dead] = 1
#         W[:, dead] = 0

#         if actorder:
#             perm = torch.argsort(torch.diag(H), descending=True)
#             W = W[:, perm]
#             H = H[perm][:, perm]

#         Losses = torch.zeros_like(W)
#         Q = torch.zeros_like(W)

#         damp = percdamp * torch.mean(torch.diag(H))
#         diag = torch.arange(self.columns)
#         H[diag, diag] += damp
#         H = torch.linalg.cholesky(H)
#         H = torch.cholesky_inverse(H)
#         H = np.array(H, dtype=np.float32)
#         H = paddle.to_tensor(H, dtype='float32')
#         H = paddle.linalg.cholesky(H, upper=True)
#         H = np.array(H, dtype=np.float32)
#         H = torch.tensor(H, dtype=torch.float32)
#         Hinv = H

#         for i1 in range(0, self.columns, blocksize):
#             i2 = min(i1 + blocksize, self.columns)
#             count = i2 - i1

#             W1 = W[:, i1:i2].clone()
#             Q1 = torch.zeros_like(W1)
#             Err1 = torch.zeros_like(W1)
#             Losses1 = torch.zeros_like(W1)
#             Hinv1 = Hinv[i1:i2, i1:i2]

#             for i in range(count):
#                 w = W1[:, i]
#                 d = Hinv1[i, i]

#                 if groupsize != -1:
#                     if (i1 + i) % groupsize == 0:
#                         weight_scale = compute_scales(W[:, (i1 + i):(i1 + i + groupsize)].t(), method=self.weight_quant_type)
#                         weight_scale /= self.quant_bits
                
#                 q = torch.clamp(torch.round(w / weight_scale), -self.quant_bits -1, self.quant_bits)
#                 q = q * weight_scale
#                 Q1[:, i] = q
#                 Losses1[:, i] = (w - q) ** 2 / d ** 2

#                 err1 = (w - q) / d
#                 W1[:, i:] -= err1.unsqueeze(1).matmul(Hinv1[i, i:].unsqueeze(0))
#                 Err1[:, i] = err1

#             Q[:, i1:i2] = Q1
#             Losses[:, i1:i2] = Losses1 / 2

#             W[:, i2:] -= Err1.matmul(Hinv[i1:i2, i2:])


#         torch.cuda.synchronize()
#         print('time %.2f' % (time.time() - tick))
#         print('error', self.layer.full_name(), torch.sum(Losses).item())
#         # print('snr loss', snr_loss())

#         if actorder:
#             invperm = torch.argsort(perm)
#             Q = Q[:, invperm]

#         Q = np.array(Q, dtype=np.float32)
#         Q = paddle.to_tensor(Q, dtype='float32')
#         param = self.layer.weight
#         Q = Q.t().cast(self.layer.weight.dtype)
#         paddle.assign(Q, output=param)
        
#         self.quantized = True
#         del H, Q, Hinv, W, Losses
#         paddle.device.cuda.empty_cache()



# def compute_scales(x, method='abs_max'):
#     x = np.array(x, dtype=np.float32)
#     x = paddle.to_tensor(x, dtype='float32')
#     if method == 'abs_max':
#         quant_scale = float(paddle.max(paddle.abs(x.flatten())))
#         quant_scale = 1e-8 if quant_scale == 0.0 else quant_scale
#     elif method == 'avg':
#         quant_scale = paddle.abs(x.reshape((x.shape[0], -1)))
#         quant_scale = paddle.mean(paddle.max(quant_scale, axis=(1)))
#     elif method == 'abs_max_channel_wise':
#         reduce_axis = tuple(
#             [i for i in range(len(x.shape)) if i != 1])
#         quant_scale = paddle.max(paddle.abs(x), axis=reduce_axis)
#         quant_scale = paddle.where(quant_scale == np.float32(0.0),
#                                       np.float32(1e-8), quant_scale)
#     quant_scale = np.array(quant_scale, dtype=np.float32)
#     quant_scale = torch.tensor(quant_scale, dtype=torch.float32)
#     return quant_scale


