# Copyright (c) 2023  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import paddle
import paddle.distributed.fleet as fleet
from .utils import compute_scales, k_means
from .metrics import mse_loss
from ...utils.log import logger

__all__ = ['PieceWiseSearch']


class PieceWiseSearch():
    def __init__(self,
                 k_piece=1,
                 bits_length=8,
                 search_piece=False,
                 search_alpha_min=0.2,
                 search_alpha_max=0.8,
                 search_scale_min=1.,
                 search_scale_max=5.,
                 weight_quant_method='abs_max_channel_wise',
                 act_quant_method='abs_max',
                 use_clip=False,
                 search_clip=False,
                 loss_function=mse_loss,
                 dp_degree=1):
        '''
        PieceWiseSearch provides to search k_piece, alpha and scale.

        Args:
        k_piece (int): Number of k pieces. Default: 1.
        bits_length (int): Number of bits to quantize the weight. Default: 8.
        search_piece (bool): Whether to search the best k piece. Default: False.
        search_alpha_min (float): Minimum alpha for search. Default: 0.2.
        search_alpha_max (float): Maximum alpha for search. Default: 0.8.
        search_scale_min (float): Minimum scale for search. Default: 1.
        search_scale_max (float): Maximum scale for search. Default: 5.
        weight_quant_method (str): Weight quantization method. Choosen from abs_max, abs_max_channel_wise and avg. Default: abs_max_channel_wise.
        act_quant_method (str): Activation quantization method. Choosen from abs_max, avg. Default: abs_max.
        loss_function (callable): Loss function. Default: mse_loss.
        '''
        self.k_piece = k_piece
        self.bits_length = bits_length
        self.search_piece = search_piece
        self.search_alpha_min = search_alpha_min
        self.search_alpha_max = search_alpha_max
        self.search_scale_min = search_scale_min
        self.search_scale_max = search_scale_max
        self.weight_quant_method = weight_quant_method
        self.act_quant_method = act_quant_method
        self.bnt = (1 << (bits_length - 1)) - 1
        self.loss_function = loss_function
        self.use_clip = use_clip
        self.search_clip = search_clip
        self.dp_degree = dp_degree
        try:
            hcg = fleet.get_hybrid_communicate_group()
            self.mp_degree = hcg.get_model_parallel_world_size()
            self.mp_id = hcg.get_model_parallel_rank()
            self.dp_id = hcg.get_data_parallel_rank()
        except:
            self.mp_degree = 1
            self.mp_id = 0
            self.dp_id = 0

    def search(self, layer_name, sampled_input, act_abs_max, weight):
        act = sampled_input
        act.stop_gradient = True
        logger.debug(f"[smooth search] search input of {layer_name}")
        dtype = weight.dtype
        origin_out = paddle.matmul(act, weight)
        w_abs_max = weight.abs().max(axis=-1, keepdim=True)
        rw_abs_max = w_abs_max.reshape(act_abs_max.shape)
        
        smooth_scale_out = None
        global_loss = float('inf')
        best_scale = None

        if self.search_clip:
            piece_range = [1] + list(range(1, self.k_piece + 1))
        else:
            piece_range = list(range(1, self.k_piece + 1))

        alpha_max = self.search_scale_max if self.search_scale_max is not None else self.search_alpha_max
        if alpha_max < 1.0:
            alpha_range = np.arange(self.search_alpha_min, alpha_max, 0.01).tolist()
        else:
            alpha_range = np.arange(self.search_alpha_min, 1.0, 0.01).tolist() + np.arange(1.0, alpha_max, 0.5).tolist()
        len_alpha_range = len(alpha_range)
        if len_alpha_range % self.dp_degree != 0:
            additional_range = [alpha_max] * (((len_alpha_range // self.dp_degree) + 1) \
                  * self.dp_degree - len_alpha_range)
            alpha_range = alpha_range + additional_range
        for k_idx, k_piece in enumerate(piece_range):
            if not self.search_piece:
                k_piece = self.k_piece
            logger.debug(f"Search {k_piece} Piece")
            centroids, labels = k_means(act_abs_max, k_piece)
            piece = ['piece_{}'.format(a) for a in range(len(centroids))]
            for i in range(len(centroids)):
                logger.debug('search for piece {}; centroids value is {}'.format(
                    piece[i], float(centroids[centroids.argsort()[i: i + 1]].cast('float32'))))
                calibration_loss = float('inf')
                final_alpha = None
                mask_for_search = paddle.where(labels == centroids.argsort()[i],
                                               1., 0.)
                mask_for_ones = paddle.where(mask_for_search == 0., 1., 0.)
                mask_for_search = mask_for_search.cast(dtype)
                mask_for_ones = mask_for_ones.cast(dtype)

                for j in range(len(alpha_range) // self.dp_degree):
                    alpha = alpha_range[self.dp_id + j * self.dp_degree]
                    alpha = round(alpha, 2)

                    if alpha < 1:
                        act_abs_max_tmp = act_abs_max.detach().clone()
                        s = paddle.clip(paddle.pow(act_abs_max_tmp, alpha) / paddle.pow(
                            rw_abs_max, 1 - alpha), min=1e-5)
                        
                        if self.use_clip or (k_piece == 1 and k_idx == 1 and self.search_clip):
                            s = paddle.clip(act_abs_max_tmp / paddle.max(act_abs_max / s), min=1)
                        del act_abs_max_tmp
                        smooth_scale = s * mask_for_search
                    else:
                        smooth_scale = paddle.to_tensor(alpha,  dtype=dtype) * mask_for_search

                    if smooth_scale_out is not None:
                        mask_for_ones_new = paddle.where(
                            smooth_scale_out == 0., 1., 0.)
                        mask_for_ones *= mask_for_ones_new.cast(dtype)
                        smooth_scale_ = smooth_scale_out + smooth_scale
                        smooth_scale_tmp = smooth_scale_ + mask_for_ones
                    else:
                        smooth_scale_tmp = smooth_scale + mask_for_ones

                    new_act = act / smooth_scale_tmp
                    new_weight = weight * smooth_scale_tmp.reshape(
                        w_abs_max.shape)

                    quant_scale = compute_scales(
                        new_act, method=self.act_quant_method)
                    quant_act = paddle.clip(
                        paddle.round(new_act / quant_scale * self.bnt),
                        -self.bnt - 1, self.bnt)
                    quant_dequant_act = quant_act / self.bnt * quant_scale

                    quant_scale = compute_scales(
                        new_weight, method=self.weight_quant_method)
                    quant_weight = paddle.clip(
                        paddle.round(new_weight / quant_scale * self.bnt),
                        -self.bnt - 1, self.bnt)
                    quant_dequant_weight = quant_weight / self.bnt * quant_scale
                    new_out = paddle.matmul(quant_dequant_act,
                                            quant_dequant_weight)

                    cur_loss = self.loss_function(origin_out, new_out)
                    if cur_loss <= calibration_loss:
                        calibration_loss = cur_loss
                        final_smooth_scale = smooth_scale
                        final_alpha = alpha

                logger.debug("Layer {} Piece {}, loss: {}, alpha : {}".format(
                    layer_name, piece[i], float(calibration_loss.cast('float32')), final_alpha))
                final_smooth_scale = final_smooth_scale.astype(dtype)
                if self.dp_degree > 1:
                    final_smooth_scale_list = []
                    calibration_loss_list = []
                    paddle.distributed.all_gather(final_smooth_scale_list, final_smooth_scale)
                    paddle.distributed.all_gather(calibration_loss_list, calibration_loss)

                    for dp_rank in range(self.dp_degree):
                        if calibration_loss_list[dp_rank * self.mp_degree + self.mp_id] < calibration_loss:
                            calibration_loss = calibration_loss_list[dp_rank * self.mp_degree + self.mp_id]
                            final_smooth_scale = final_smooth_scale_list[dp_rank * self.mp_degree + self.mp_id]

                if smooth_scale_out is None:
                    smooth_scale_out = final_smooth_scale
                else:
                    smooth_scale_out += final_smooth_scale

            if self.dp_degree > 1:
                smooth_scale_out_list = []
                calibration_loss_list = []

                paddle.distributed.all_gather(smooth_scale_out_list, smooth_scale_out)
                paddle.distributed.all_gather(calibration_loss_list, calibration_loss)

                for i in range(self.dp_degree):
                    if calibration_loss_list[i * self.mp_degree + self.mp_id] < calibration_loss:
                        calibration_loss = calibration_loss_list[i * self.mp_degree + self.mp_id]
                        smooth_scale_out = smooth_scale_out_list[i * self.mp_degree + self.mp_id]

            if calibration_loss < global_loss:
                global_loss = calibration_loss
                best_scale = smooth_scale_out
                if self.search_piece:
                    logger.debug(f"Find Better K-Piece {k_piece}")
            else:
                if best_scale is not None:
                    logger.debug("Early Stopped")
                    break
            if not self.search_piece:
                break
        
        return best_scale


