""" only use for api4 model """
import paddle
from paddle import nn
from paddle.distributed.fleet.layers.mpu import mp_ops
from paddle.nn import functional as F
from paddle.nn.quant.format import ConvertibleQuantedLayer
from paddleslim.quant.quanters.channel_wise_abs_max import FakeQuanterChannelWiseAbsMaxObserverLayer
from ernie_core.models.sequence_parallel_utils import (
    AllGatherOp,
    ColumnSequenceParallelLinear,
    ReduceScatterOp,
    RowSequenceParallelLinear,
    mark_as_sequence_parallel_parameter,
)
from paddle import ParamAttr
from paddle.nn.initializer import Constant


class RowSequenceParallelQuantedLoRALinear(ConvertibleQuantedLayer):
    """Row sequence parallel linear"""
    def __init__(self, layer: nn.Layer, q_config):
        """Initialize."""
        super().__init__()
        
        self.weight = layer.weight
        self.lora_A = layer.lora_A
        self.lora_B = layer.lora_B
        self.scaling = layer.scaling
        self.bias = layer.bias
        self.name = layer.name
        self.model_parallel_group = layer.model_parallel_group
        self.input_is_parallel = layer.input_is_parallel
        # Mark the weight as unmerged
        self.merged = False
        self.merge_weights = layer.merge_weights

        self.lora_A.is_distributed = True
        self.lora_A.split_axis = 0
        self.lora_B.is_distributed = False
        mark_as_sequence_parallel_parameter(self.lora_B)
        
        # Freezing the pre-trained weight matrix
        self.weight.stop_gradient = True
        
        # For FakeQuant
        self.weight_quanter = None
        self.activation_quanter = None
        if q_config.weight is not None:
            self.weight_quanter = q_config.weight._instance(layer)
        if q_config.activation is not None:
            self.activation_quanter = q_config.activation._instance(layer)
        self._base_weight_scale = self.weight.abs().max(axis=0)
        
        # ### add
        # self.shift_weight = self.create_parameter(
        #     shape=[self.weight.shape[0]],
        #     attr=ParamAttr(initializer=Constant(value=0.)),
        #     dtype=self.weight.dtype)
        # self.shift_weight.stop_gradient = True
        # self.shift_weight.is_distributed = True
        # ### multiply
        # self.smooth_weight = self.create_parameter(
        #     shape=[self.weight.shape[0]],
        #     attr=ParamAttr(initializer=Constant(value=1.)),
        #     dtype=self.weight.dtype)
        # self.smooth_weight.stop_gradient = True
        # self.smooth_weight.is_distributed = True
    
    def _quant_dequant(self, x, scale, quant_bit=8):
        """ Fake quant.
        """
        def round_ste(x):
            return (x.round() - x).detach() + x

        qmax = 2**(quant_bit - 1) - 1
        qmin = -2**(quant_bit - 1)
        s = scale / qmax
        quant_x = paddle.clip(round_ste(x / s), qmin, qmax)
        dequant_x = s * quant_x
        return dequant_x
    
    def forward(self, input):
        """Forward function."""
        quant_input = self.activation_quanter(input) if self.activation_quanter is not None else input
        quant_weight = self._quant_dequant(self.weight, self._base_weight_scale)
        return self._linear_forward(quant_input, quant_weight)
    
    def _linear_forward(self, input, weight):
        """Forward function of Linear."""
        if not self.input_is_parallel:
            input_mp = mp_ops._c_split(input, group=self.model_parallel_group)
        else:
            input_mp = input
        output_parallel = F.linear(x=input_mp, weight=weight, name=self.name)
        output_ = ReduceScatterOp.apply(output_parallel)
        result_mp = output_ + self.bias if self.bias is not None else output_

        if not self.merged:
            input_mp = input_mp @ self.lora_A
            input_mp = ReduceScatterOp.apply(input_mp)
            delta_mp = (input_mp @ self.lora_B) * self.scaling
            result_mp += delta_mp
        return result_mp

    def train(self):
        """Set module to training mode."""
        super().train()
        if self.merge_weights and self.merged:
            # Make sure that the weights are not merged
            new_weight = self.weight - self.lora_A @ self.lora_B * self.scaling
            self.weight.set_value(new_weight)
            self.merged = False

    def eval(self):
        """Set module to evaluation mode."""
        super().eval()
        if self.merge_weights and not self.merged:
            # Merge the weights and mark it
            new_weight = self.weight + self.lora_A @ self.lora_B * self.scaling
            self.weight.set_value(new_weight)
            self.merged = True

    def weights_to_quanters(self):
        """weights_to_quanters"""
        return [("weight", "weight_quanter")]

    def activation_quanters(self):
        """activation_quanters"""
        return ["activation_quanter"]



class ColumnSequenceParallelQuantedLoRALinear(ConvertibleQuantedLayer):
    """Column sequence parallel linear"""
    def __init__(self, layer: nn.Layer, q_config):
        """Initialize."""
        super().__init__()
        self.weight = layer.weight
        self.lora_A = layer.lora_A
        self.lora_B = layer.lora_B
        self.scaling = layer.scaling
        self.bias = layer.bias
        self.name = layer.name
        self.is_mp = layer.is_mp
        self.model_parallel_group = layer.model_parallel_group
        self.gather_output = layer.gather_output

        # Mark the weight as unmerged
        self.merged = False
        self.merge_weights = layer.merge_weights

        self.lora_A.is_distributed = False
        mark_as_sequence_parallel_parameter(self.lora_A)
        self.lora_B.is_distributed = True
        self.lora_B.split_axis = 1
        
        # Freezing the pre-trained weight matrix
        self.weight.stop_gradient = True
        
        # For FakeQuant
        self.weight_quanter = None
        self.activation_quanter = None
        if q_config.weight is not None:
            self.weight_quanter = q_config.weight._instance(layer)
        if q_config.activation is not None:
            self.activation_quanter = q_config.activation._instance(layer)
        self._base_weight_scale = self.weight.abs().max(axis=0)
        
        # ### add
        # self.shift_weight = self.create_parameter(
        #     shape=[self.weight.shape[0]],
        #     attr=ParamAttr(initializer=Constant(value=0.)),
        #     dtype=self.weight.dtype)
        # self.shift_weight.stop_gradient = True
        # self.shift_weight.is_distributed = True
        # ### multiply
        # self.smooth_weight = self.create_parameter(
        #     shape=[self.weight.shape[0]],
        #     attr=ParamAttr(initializer=Constant(value=1.)),
        #     dtype=self.weight.dtype)
        # self.smooth_weight.stop_gradient = True
        # self.smooth_weight.is_distributed = True
            
    def _quant_dequant(self, x, scale, quant_bit=8, only_dequant=False):
        """ Fake quant.
        """
        def round_ste(x):
            return (x.round() - x).detach() + x
        
        qmax = 2**(quant_bit - 1) - 1
        qmin = -2**(quant_bit - 1)
        s = scale / qmax
        if not only_dequant:
            quant_x = paddle.clip(round_ste(x / s), qmin, qmax)
        else:
            quant_x = x
        dequant_x = s * quant_x
        return dequant_x
    
    def forward(self, input):
        """Forward function."""
        quant_input = self.activation_quanter(input) if self.activation_quanter is not None else input
        quant_weight = self._quant_dequant(self.weight, self._base_weight_scale)
        return self._linear_forward(quant_input, quant_weight)
            
    def _linear_forward(self, input, weight):
        """Forward function of Linear."""
        if self.is_mp:
            input_parallel = AllGatherOp.apply(input)
        else:
            input_parallel = input
        result_mp = F.linear(x=input_parallel, weight=weight, bias=self.bias, name=self.name)

        if not self.merged:
            input_a = input @ self.lora_A
            input_a = AllGatherOp.apply(input_a)
            delta_mp = (input_a @ self.lora_B) * self.scaling
            result_mp += delta_mp

        if self.gather_output and self.is_mp:
            result = mp_ops._c_concat(result_mp, group=self.model_parallel_group)
        else:
            result = result_mp
        return result

    def train(self):
        """Set module to training mode."""
        super().train()
        if self.merge_weights and self.merged:
            # Make sure that the weights are not merged
            new_weight = self.weight - self.lora_A @ self.lora_B * self.scaling
            self.weight.set_value(new_weight)
            self.merged = False

    def eval(self):
        """Set module to evaluation mode."""
        super().eval()
        if self.merge_weights and not self.merged:
            # Merge the weights and mark it
            new_weight = self.weight + self.lora_A @ self.lora_B * self.scaling
            self.weight.set_value(new_weight)
            self.merged = True

    def weights_to_quanters(self):
        """weights_to_quanters"""
        return [("weight", "weight_quanter")]

    def activation_quanters(self):
        """activation_quanters"""
        return ["activation_quanter"]
