# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from ...lc.quantizers.quant_func import quantize_nf4, dequantize_nf4

class NF4QuanterDequanter(paddle.nn.Layer):
    """
    NF4 量化操作。
    """
    def __init__(self, quanter, dequanter):
        """
        用于处理消息队列中的元素。
        
         Args:
             quanter (Quanter): 用于获取和放回消息的对象。
             dequanter (Dequanter): 用于删除消息队列中的元素的对象。
        
        Returns:
            None.
        
        """
        super().__init__()
        self._quanter = quanter
        self._dequanter = dequanter

    def forward(self, input):
        """
        将输入数据转换为量化后的输出
        
        Args:
            input (torch.Tensor): 原始输入数据，维度是[batch_size, channels]
        
        Returns:
            torch.Tensor: 量化后的输出数据，维度是[batch_size, channels]
        
        """
        out = input
        if self._quanter is not None:
            out = self._quanter(out)
        if self._dequanter is not None:
            out = self._dequanter(out)
        return out

    @staticmethod
    def from_quanter(quanter):
        """
        从 quanter 转换为 QuanterDequanter。
        
        Args:
            quanter (Optional[Union[NF4Quanter, NF4Dequanter]]):
                表示 Quanter 或 Dequanter 的对象。
        
        Returns:
            Optional[QuanterDequanter]: 如果 `quanter` 不为 None，则返回其对应的 QuanterDequanter 对象；否则返回 None。
        
        Raises:
            AssertionError: 如果 `quanter` 为 None。
        """
        assert quanter is not None
        return NF4QuanterDequanter(
            NF4Quanter.from_quanter(quanter),
            NF4Dequanter.from_quanter(quanter),
        )

class NF4Quanter(paddle.nn.Layer):
    def __init__(self, scales, zero_point=None, quant_axis=None, bit_length=4, blocksize=4096):
        """
        用于初始化量化操作的基类。
        
        Args:
            scales (np.ndarray[float32] | Tensor): 需要进行量化处理的张量。如果scales是一个numpy数组，则将其转换为Tensor类型。
            zero_point (Optional[np.ndarray[int32] | Tensor]): 量化的零点。默认为None，表示没有零点。如果不是None，它必须是一个numpy数组或Tensor类型。
            quant_axis (Optional[int]): 量化操作的轴序号。默认为-1，表示沿着最后一维度进行量化。
            bit_length (Optional[int]): 表示量化的位宽。默认值为4。
            blocksize (Optional[int]): 每个块的大小。默认为4096。
        
        Returns:
            None: 初始化量化操作的基类实例。
        """
        super().__init__()
        self._scales = paddle.to_tensor(scales, dtype="float32")
        self._zero_point = (
            paddle.zeros([1], dtype="float32")
            if zero_point is None
            else paddle.to_tensor(zero_point)
        )
        self._quant_axis = -1 if quant_axis is None else quant_axis
        self._bit_length = bit_length
        self._blocksize = blocksize

    def forward(self, input):
        """
        Args:
            input (Tensor): Input Tensor.
        
        Returns:
            Tensor：输出Tensor。
        
        """
        _scales = paddle.cast(self._scales, input.dtype)
        _zero_point = paddle.cast(self._zero_point, input.dtype)
        # _scales = self._scales
        # _zero_point = self._zero_point
        return dequantize_nf4(input, _scales, self._blocksize).cast(input.dtype)

    @staticmethod
    def from_quanter(quanter):
        """
        从Quanter实例创建NF4Quanter对象。
        
        Args:
            quanter (paddleflow.quantization.Quanter): 待转换的Quanter实例。
        
        Returns:
            paddleflow.quantization.NF4Quanter: 转换后的NF4Quanter实例。
        """
        return NF4Quanter(
            quanter.scales(),
            zero_point=quanter.zero_points(),
            quant_axis=quanter.quant_axis(),
            bit_length=quanter.bit_length(),
            blocksize=quanter.blocksize(),
        )

class NF4Dequanter(paddle.nn.Layer):
    def __init__(self, scales, zero_point=None, quant_axis=None, bit_length=4, blocksize=4096):
        """
        用于初始化量化操作的基类。
        
        Args:
            scales (np.ndarray[float32] | Tensor): 需要进行量化处理的张量。如果scales是一个numpy数组，则将其转换为Tensor类型。
            zero_point (Optional[np.ndarray[int32] | Tensor]): 量化的零点。默认为None，表示没有零点。如果不是None，它必须是一个numpy数组或Tensor类型。
            quant_axis (Optional[int]): 量化操作的轴序号。默认为-1，表示沿着最后一维度进行量化。
            bit_length (Optional[int]): 表示量化的位宽。默认值为4。
            blocksize (Optional[int]): 每个块的大小。默认为4096。
        
        Returns:
            None: 初始化量化操作的基类实例。
        """
        super().__init__()
        self._scales = paddle.to_tensor(scales, dtype="float32")
        self._zero_point = (
            paddle.zeros([1], dtype="float32")
            if zero_point is None
            else paddle.to_tensor(zero_point)
        )
        self._quant_axis = -1 if quant_axis is None else quant_axis
        self._bit_length = bit_length
        self._blocksize = blocksize

    def forward(self, input):
        """
        Args:
            input (Tensor): 输入的 Tensor，形状为 [batch_size, channels, height, width]。
        
        Returns:
            Tensor: 输出的 Tensor，形状为 [batch_size, channels, height, width]，数据类型为 bfloat16。
        
        """
        _scales = paddle.cast(self._scales, "float32")
        # _scales = self._scales
        # _zero_point = self._zero_point
        origin_shape = input.shape
        input = input.reshape([-1])
        return dequantize_nf4(input, _scales, self._blocksize).cast("bfloat16").reshape([-1, origin_shape[-1]])

    @staticmethod
    def from_quanter(quanter):
        """
        从Quanter对象中创建NF4Dequanter对象
        
        Args:
            quanter (Quanter): 待转换的Quanter对象
        
        Returns:
            NF4Dequanter: 返回转换后的NF4Dequanter对象
        
        """
        return NF4Dequanter(
            quanter.scales(),
            zero_point=quanter.zero_points(),
            quant_axis=quanter.quant_axis(),
            bit_length=quanter.bit_length(),
            blocksize=quanter.blocksize(),
        )
