# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import paddle
from paddle.nn.initializer import Constant
from paddle.utils import unique_name
from paddle.framework import ParamAttr
from paddle.quantization.factory import QuanterFactory
from .base_fake_quanter import BaseFakeQuanterLayer
from .channel_wise_abs_max import CHANNEL_AXIS


def round_ste(x):
    """
    Implement Straight-Through Estimator for rounding operation.
    """
    return (x.round() - x).detach() + x


class ABQuanter(QuanterFactory):
    r"""
    Weight quantizer.
    Args:
        bit_length(int): Number of bits to represent an quantized integer in binary.
        dtype(str): data type.
        name(str): the name of the layer.
    Examples:
       .. code-block:: python
            from paddle.quantization import QuantConfig
            from paddleslim.quant.quanters import ABQuanter
            from paddle.quantization.quanters.abs_max import FakeQuanterWithAbsMaxObserver
            weight_quanter = ABQuanter(channel_wise=True)
            act_quanter = FakeQuanterWithAbsMaxObserver()
            q_config = QuantConfig(activation=act_quanter, weight=weight_quanter)
    """
    def __init__(self,
                 bit_length=8,
                 channel_wise=True,
                 sign=True,
                 dtype='float32',
                 name=None,
                 windows=None,
                 quanter=None,
                 fix_alpha=0.5):
        """Initializes the `Quantizer` class with specific parameters."""
        super().__init__(
            bit_length=bit_length,
            channel_wise=channel_wise,
            sign=sign,
            dtype=dtype,
            name=name,
            windows=windows,
            quanter=quanter,
            fix_alpha=fix_alpha)

    def _get_class(self):
        """ Return ABQuanterLayer
        """
        return ABQuanterLayer


class ABQuanterLayer(BaseFakeQuanterLayer):
    r"""
    ABQuanterLayer.
    """
    def __init__(self,
                 layer,
                 bit_length=8,
                 channel_wise=False,
                 sign=True,
                 dtype='float32',
                 name=None,
                 windows=None,
                 quanter=None,
                 fix_alpha=None):
        """Initializes ABQuanterLayer."""
        super().__init__()
        self._bit_length = bit_length
        self._sign = sign
        self._qmin, self._qmax = self.qmin_qmax
        self._current_iters = -1
        self._windows = windows if windows is not None else [0, 0]
        self._fix_alpha = fix_alpha
        if quanter:
            self._quanter = quanter._instance(layer)
            self._scale = self._quanter.scales()
            self._quant_axis = self._quanter.quant_axis()
        else:
            self._init_scale(layer, channel_wise, name, dtype)
            self._quanter = None

    def _init_scale(self, layer, channel_wise, name, dtype):
        """ Init scale.
        """
        if channel_wise:
            for key in CHANNEL_AXIS.keys():
                if issubclass(type(layer), key):
                    self._quant_axis = CHANNEL_AXIS[key]
                    break
            self._channel_num = layer.weight.shape[self._quant_axis]
        else:
            self._quant_axis = -1
            self._channel_num = 1

        scale_prefix = "{}.scale".format(name) if name else 'quant_dequant.scale'
        self._scale_name = unique_name.generate(scale_prefix)
        scale_attr = ParamAttr(
            name=self._scale_name,
            initializer=Constant(0.001),
            trainable=False, )
        self._scale = self.create_parameter(
            shape=[self._channel_num], attr=scale_attr, dtype=dtype)
        self._scale.stop_gradient = True

    def forward(self, inputs):
        """
        forward function.
        """
        if self.training:
            alpha = self._update_params(inputs.detach())
            if self._quanter is None:
                if self._fix_alpha is not None or self._current_iters <= self._windows[1]:
                    with paddle.no_grad():
                        qdq_inputs = self._quant_dequant(inputs)
                else:
                    qdq_inputs = self._quant_dequant(inputs, update=True)
                return inputs * (1 - alpha) + alpha * qdq_inputs
            else:
                if self._current_iters <= self._windows[1]:
                    with paddle.no_grad():
                        qdq_inputs = self._quanter(inputs.detach())
                else:
                    qdq_inputs = self._quanter(inputs)
                return inputs * (1 - alpha) + alpha * qdq_inputs
        else:
            if self._quanter is None:
                return self._quant_dequant(inputs)
            else:
                return self._quanter(inputs)

    def _update_params(self, inputs):
        """ Update scale and alpha.
        """
        if self._quanter is None:
            if self._channel_num > 1: # for weight
                reduce_axis = tuple([
                    i for i in range(len(inputs.shape)) if i != self._quant_axis
                ])
                abs_max_values = paddle.max(
                    paddle.abs(inputs), axis=reduce_axis)
                self._scale.set_value(abs_max_values)

        if self._fix_alpha:
            return self._fix_alpha

        t0, t1 = self._windows
        self._current_iters += 1
        if t0 == t1 or self._current_iters > t1:
            return 1.
        if self._current_iters <= t0:
            return 0.
        if self._current_iters <= t1:
            return 1 - math.pow((t1 - self._current_iters) / (t1 - t0), 3)

    def _quant_dequant(self, x, update=False):
        """ Fake quant.
        """
        s = self._scale / self._qmax

        quant_x = paddle.clip(round_ste(x / s), self._qmin, self._qmax)
        dequant_x = s * quant_x
        return dequant_x

    def bit_length(self):
        """ Return the bit length of quantized data.
        """
        return self._bit_length

    def quant_axis(self):
        """ Return quantization axis.
        """
        return self._quant_axis

    def scales(self):
        """ Return output scales.
        """
        return self._scale

    def zero_points(self):
        """ Return output zero points.
        """
        return None


