# Copyright (c) 2023 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import numpy as np
import matplotlib.pyplot as plt

from .log import logger

class AutoDataSelection(object):
    """
        1. get embeddings of each query
        2. get average cosine similarity
        3. cluster embeddings by cosine similarity
        4. select one query from each cluster
        5. save or load embeddings if need

        examples:
            for file in files:
                search_data = AutoDataSelection(model, ErnieBotEmbeddings)
                ptq_dials = []
                path = os.path.join(args.ptq_file, file)
                with open(path, "r") as fin:
                    for line in fin:
                        line = json.loads(line.strip())
                        ptq_dials.append(line)
                for ptq_dial in ptq_dials:
                    search_query = [{"role": "user", "utterance": [ptq_dial[0]["utterance"]]}]
                    predictor.predict_ptq(search_query)
                    search_data.collect_embeddings(ptq_dial)
                clusters = search_data.cluster_data_by_embeddings()
                target_list = search_data.get_cluster_data_by_embeddings(clusters)
    """
    def __init__(self, model, layer_type):
        self.model = model
        self.forward_post_hook_handle = None
        self.embeddings_list = []
        self.layer_type = layer_type
        self._apply_hook()
    
    def _apply_hook(self):
        """
            Apply forward post-processing hook to all instances of a given Embedding layer type.
            
            Args:
                None
            
            Returns:
                Nothing
        """
        if type(self.model) == self.layer_type:
            self.forward_post_hook_handle = self.model.register_forward_post_hook(self._forward_post_hook)
    
    def _forward_post_hook(self, layer, input, output):
        """
            Handle the forward propagation process.         
        """
        self.embeddings = output[0]
        return output
    
    def collect_embeddings(self, inputs, dtype='float16', max_seq_len=3072):
        """
            Collect embeddings of each query. 
        """
        embeddings = paddle.mean(self.embeddings, axis=-1)
        embeddings = paddle.cast(embeddings, dtype=dtype)
        embeddings = self.pad_embeddings(embeddings, dtype=dtype, max_seq_len=max_seq_len)
        self.embeddings_list.append({"inputs": inputs, "embeddings": embeddings})
    
    def pad_embeddings(self, insts, pad_id=0, max_seq_len=3072, dtype='float16'):
        """
            Pad the instances to the max sequence length.
        """
        padding_insts = []
        for inst in insts:
            if len(inst) >= max_seq_len:
                inst_data = inst[:max_seq_len]
            else:
                pad_data = paddle.to_tensor([pad_id] * (max_seq_len - len(inst)), dtype=dtype)
                inst_data = paddle.to_tensor(list(inst) + list(pad_data), dtype=dtype)
            inst_data = paddle.squeeze(inst_data)
            padding_insts.append(inst_data)
        return paddle.to_tensor(padding_insts, dtype=dtype)

    def cluster_data_by_embeddings(self, plot_similarity=False, save_name="", target_iters=None):
        """
            Cluster embeddings by cosine similarity.
        """
        clusters = []
        cos_sim_func = paddle.nn.CosineSimilarity(axis=1)
        max_similarity_list = []
        if target_iters is None:
            target_iters = len(self.embeddings_list)
        with paddle.no_grad():
            for idx in range(0, len(self.embeddings_list)):
                dials_embeddings = self.embeddings_list[idx]
                embeddings = dials_embeddings["embeddings"]
                if idx == 0:
                    avg_embeddings = paddle.mean(embeddings, axis=0, keepdim=True)
                    avg_similarity = 0
                else:
                    sims = cos_sim_func(embeddings.cast("float32"), avg_embeddings.cast("float32"))
                    sims = float(paddle.mean(sims))
                    avg_similarity = (avg_similarity * (idx - 1) + sims) / idx
                    embeddings = paddle.mean(embeddings, axis=0, keepdim=True)
                    avg_embeddings = (avg_embeddings * idx + embeddings) / (idx + 1)
                if idx > 50:
                    break
            logger.debug(f"similarity threshold: {avg_similarity}")

            for idx in range(0, len(self.embeddings_list)):
                dials_embeddings = self.embeddings_list[idx]
                inputs = dials_embeddings["inputs"]
                embeddings = dials_embeddings["embeddings"]

                if idx == 0:
                    embeddings = paddle.mean(embeddings, axis=0, keepdim=True)
                    clusters.append({"centroid": embeddings, "embeddings": [embeddings], "inputs": [inputs]})
                    continue

                max_similarity = -1
                best_cluster = None

                for cluster in clusters:
                    sims = cos_sim_func(embeddings.cast("float32"), cluster["centroid"].cast("float32"))
                    sims = float(paddle.mean(sims))
                    if sims >= max_similarity:
                        max_similarity = sims
                        best_cluster = cluster
                
                logger.debug(f"max similarity: {max_similarity}")
                max_similarity_list.append(max_similarity)
                embeddings = paddle.mean(embeddings, axis=0, keepdim=True)
                if max_similarity < avg_similarity:
                    clusters.append({"centroid": embeddings, "embeddings": [embeddings], "inputs": [inputs]})
                else:
                    best_cluster["embeddings"].append(embeddings)
                    best_cluster["inputs"].append(inputs)
                    for i in range(embeddings.shape[0]):
                        best_cluster["centroid"][i] = (
                            best_cluster["centroid"][i] * paddle.to_tensor(
                                len(best_cluster["embeddings"])) + embeddings[i]) / paddle.to_tensor((
                                    len(best_cluster["embeddings"]) + 1))
                if idx == target_iters:
                    break
                if idx % 5 == 0:
                    logger.debug(f"[Cluster] Iteration {idx} of {len(self.embeddings_list)}")

        for cluster in clusters:
            if len(cluster["inputs"]) > 1:
                logger.debug(f"len of cluster: {len(cluster['inputs'])}")
        
        if plot_similarity:
            plt.plot(np.array(max_similarity_list), '*')
            save_fig_path = str(save_name) + '.png'
            plt.savefig(save_fig_path)
            plt.close()

        self._remove_hook()
        return clusters
    
    def get_cluster_data_by_embeddings(self, clusters):
        """
            For each cluster, append the most similar instance to target list.
        """
        target_list = []
        cos_sim_func = paddle.nn.CosineSimilarity(axis=1)
        for cluster in clusters:
            sims = []
            if len(cluster["embeddings"]) == 1:
                target_list.extend(cluster["inputs"][0])
            else:
                centroid = cluster["centroid"]
                for embedding in cluster["embeddings"]:
                    sim = cos_sim_func(embedding, centroid)
                    sims.append(float(sim))
                sim_max = max(sims)
                index_max = sims.index(sim_max)
                target_list.extend(cluster["inputs"][index_max])
        return target_list

    def _remove_hook(self):
        self.forward_post_hook_handle.remove()
        self.forward_post_hook_handle = None
